# coding: utf-8
# pylint: disable=protected-access
from itertools import imap
from types import NoneType

from django.core.exceptions import FieldDoesNotExist
from django.db.models.fields import AutoField
from django.db.models.fields.files import FileField
from django.db.models.fields.related import ManyToManyField
from django.db.models.fields.related import RelatedField
from django.db.models.fields.reverse_related import ForeignObjectRel
from django.db.models.fields.reverse_related import OneToOneRel
from django.db.models.manager import Manager
from django.utils.encoding import force_text
from django.utils.encoding import force_unicode
from m3_django_compat import ModelOptions

from educommon.django.db.utils import LazyModel
from educommon.report.constructor.utils import get_field
from educommon.utils.misc import cached_property

from .utils import get_columns_hierarchy


def skip_field(model, field):
    u"""Возвращает True, если поле модели должно игнорироваться конструктором.

    Игнорируются:
        - поля с типами ``AutoField`` и ``FileField``;
        - поля указанные в ``model.report_constructor_params['except']``;
        - поля *не* указанные в ``model.report_constructor_params['only']``;
        - поля моделей, у которых ``model.report_constructor_params['skip']``
          содержит истинное значение;

    :rtype: bool
    """
    if isinstance(field, (AutoField, FileField)):
        return True

    if not hasattr(model, 'report_constructor_params'):
        return False

    params = model.report_constructor_params
    assert isinstance(params, dict), params
    assert sum(
        1 if param in params else 0
        for param in ('skip', 'except', 'only')
    ) == 1, (u'Должен быть указан только один параметр.', model, params)

    if 'skip' in params and params['skip']:
        return True

    if isinstance(field, ForeignObjectRel):
        accessor_name = field.get_accessor_name()
    else:
        accessor_name = field.name

    if 'except' in params and accessor_name in params['except']:
        return True

    if 'only' in params and accessor_name not in params['only']:
        return True

    return False


class ColumnDescriptor(object):

    u"""Дескриптор поля модели, входящего в состав колонки отчета.

    Предназначен для извлечения из моделей необходимой информации.
    """

    def __init__(self, parent, field, level=0):
        u"""Инициализация экземпляра класса.

        :param parent: Дескриптор верхнего уровня. Определяет дескриптор для
            поля, из которого модель ссылается на поле данного дейскриптора.
        :type parent: ColumnDescriptor or NoneType

        :param field: Поле модели, соответствующее данному дескриптору.

        :param int level: Уровень вложенности дескриптора.
        """
        assert isinstance(parent, (ColumnDescriptor, NoneType)), parent

        self.parent = parent
        self.field = field
        self.level = level
        self.children = []

        if self.parent:
            self.parent.children.append(self)

    @staticmethod
    def create(model, field_name, parent=None, level=0):
        model_params = getattr(model, 'report_constructor_params', {})
        assert sum(
            1 if param in model_params else 0
            for param in ('skip', 'except', 'only')
        ) <= 1, (
            u'Может быть указан только один параметр.', model, model_params
        )

        attr_name, _, nested_attr = field_name.partition('.')

        for field in getattr(model, '_meta').get_fields():
            if skip_field(model, field):
                continue

            if isinstance(field, ForeignObjectRel):
                accessor_name = field.get_accessor_name()
                related_model = field.related_model
            else:
                accessor_name = field.name
                if isinstance(field, RelatedField):
                    related_model = field.rel.to
                else:
                    related_model = None

            if accessor_name == attr_name:
                column_field = field
                break
        else:
            raise FieldDoesNotExist(field_name)

        result = ColumnDescriptor(parent, column_field, level)
        if nested_attr:
            if not related_model:
                raise FieldDoesNotExist(field_name)

            try:
                result = ColumnDescriptor.create(
                    related_model, nested_attr, result, level + 1
                )
            except FieldDoesNotExist as error:
                raise FieldDoesNotExist('.'.join((attr_name, str(error))))

        return result

    @property
    def accessor_name(self):
        if isinstance(self.field, ForeignObjectRel):
            result = self.field.get_accessor_name()
        else:
            result = self.field.name

        return force_unicode(result)

    @property
    def full_accessor_name(self):
        if self.parent:
            return '.'.join(
                (self.parent.full_accessor_name, self.accessor_name)
            )
        else:
            return self.accessor_name

    def __repr__(self, *args, **kwargs):
        return '{}<{}>'.format(
            self.__class__.__name__, self.full_accessor_name
        )

    @property
    def title(self):
        if isinstance(self.field, ForeignObjectRel):
            options = getattr(self.field.related_model, '_meta')
            if isinstance(self.field, OneToOneRel):
                result = options.verbose_name
            else:
                result = options.verbose_name_plural
        else:
            result = self.field.verbose_name

        return force_unicode(result)

    def is_root(self):
        return not self.parent

    @property
    def root(self):
        result = self

        while result.parent:
            result = result.parent

        return result

    def has_nested_columns(self):
        if isinstance(self.field, ForeignObjectRel):
            model = self.field.related_model
        elif isinstance(self.field, RelatedField):
            model = self.field.rel.to
        else:
            return False

        for field in getattr(model, '_meta').get_fields():
            if not skip_field(model, field):
                return True

        return False

    def get_nested_columns(self):
        if isinstance(self.field, ForeignObjectRel):
            model = self.field.related_model
        elif isinstance(self.field, RelatedField):
            model = self.field.rel.to
        else:
            return ()

        return (
            ColumnDescriptor(self, field, level=self.level + 1)
            for field in getattr(model, '_meta').get_fields()
            if not skip_field(model, field)
        )


class ModelDataSource(object):

    def __init__(self, model):
        self._model = LazyModel(model)
        self.column_descriptors_cache = {}

    @cached_property
    def model(self):
        return self._model.get_model()

    @cached_property
    def _options(self):
        return ModelOptions(self._model.get_model())

    @cached_property
    def name(self):
        u"""Внутрисистемное имя источника данных."""
        result = '.'.join((self._model.app_label, self._model.model_name))

        return force_unicode(result)

    @cached_property
    def title(self):
        u"""Видимое пользователем наименование ИД."""
        model = self._model.get_model()
        result = getattr(model, '_meta').verbose_name
        return force_unicode(result)

    def get_available_columns(self):
        model = self._model.get_model()
        return (
            ColumnDescriptor(None, field)
            for field in getattr(model, '_meta').get_fields()
            if not skip_field(model, field)
        )

    def get_column_descriptor(self, accessor_name):
        level_names = []
        model = self.model
        result = None
        for level_name in accessor_name.split('.'):
            level_names.append(level_name)
            full_accessor_name = '.'.join(level_names)

            if not model:
                raise FieldDoesNotExist(full_accessor_name)
            # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
            # Поиск поля в модели.

            for field in getattr(model, '_meta').get_fields():
                if skip_field(model, field):
                    continue

                if isinstance(field, ForeignObjectRel):
                    field_name = field.get_accessor_name()
                    field_model = field.related_model
                else:
                    field_name = field.name
                    if isinstance(field, RelatedField):
                        field_model = field.rel.to
                    else:
                        field_model = None

                if field_name == level_name:
                    column_field = field
                    model = field_model
                    break
            else:
                raise FieldDoesNotExist(
                    '{}.{}'.format(model.__name__, full_accessor_name)
                )
            # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
            # Создание дескриптора колонки, если его нет в кэше.

            if full_accessor_name in self.column_descriptors_cache:
                result = self.column_descriptors_cache[full_accessor_name]
            else:
                result = ColumnDescriptor(
                    result, column_field, len(level_names) - 1
                )
                self.column_descriptors_cache[full_accessor_name] = result
            # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        return result

    def get_data(self, *columns):
        u"""Возвращает данные для отчета.

        :rtype: tuple
        """

        def extract(data_gen):
            result = tuple(data_gen)
            if not result:
                raise StopIteration()
            elif len(result) == 1:
                return result[0]
            else:
                return result

        def get_column_count(columns):
            # Для иерархии колонок возвращает количество занимаемых столбцов.
            if columns:
                return sum(imap(get_column_count, columns.itervalues()))
            else:
                return 1

        def get_field_display(field, field_value):
            for value, name in field.flatchoices:
                if value == field_value:
                    return force_text(name, strings_only=True)

        def get_object_data(obj, attr_name, nested):
            field = None
            if attr_name:
                field = get_field(obj, attr_name)
                model = field.model
                if isinstance(field, ForeignObjectRel):
                    descriptor = getattr(model, field.get_accessor_name())
                elif isinstance(field, RelatedField):
                    descriptor = getattr(model, field.name)
                else:
                    descriptor = None

                if descriptor:
                    try:
                        attr_value = getattr(obj, attr_name)
                    except descriptor.RelatedObjectDoesNotExist:
                        attr_value = None
                else:
                    attr_value = getattr(obj, attr_name)
            else:
                attr_value = obj
                descriptor = None

            # Обратная связь или M2M. Такие данные выводятся в виде кортежа.
            if (
                attr_name and
                (
                    isinstance(field, ForeignObjectRel) and
                    hasattr(descriptor, 'related_manager_cls') and
                    isinstance(attr_value, descriptor.related_manager_cls) or
                    isinstance(field, ManyToManyField) and
                    isinstance(attr_value, Manager)
                )
            ):
                objects = attr_value.all()
                if objects:
                    yield tuple(
                        extract(get_object_data(related_obj, None, nested))
                        for related_obj in objects
                    )
                else:
                    yield ((None,) * get_column_count(nested),)

            # Прямая связь (ForeignKey или OneToOneField).
            elif nested:
                if attr_value:
                    for k, v in nested.iteritems():
                        for object_data in get_object_data(attr_value, k, v):
                            yield object_data
                else:
                    for _ in xrange(get_column_count(nested)):
                        yield None

            # Поле с данными.
            else:
                if field.flatchoices:
                    attr_value = get_field_display(field, attr_value)

                yield attr_value

        def get_rows_data(hierarchy):
            for obj in self.model.objects.all():
                row_data = tuple(get_object_data(obj, None, hierarchy))
                if row_data and any(row_data):
                    yield row_data

        columns_hierarchy = get_columns_hierarchy(*columns)

        return tuple(get_rows_data(columns_hierarchy))
